package Net::DNS::Nameserver;

use strict;
use warnings;

our $VERSION = (qw$Id: Nameserver.pm 1949 2023-11-27 11:51:17Z willem $)[2];



use integer;
use Carp;
use Net::DNS;
use Net::DNS::ZoneFile;

use IO::Select;
use IO::Socket::IP 0.38;
use IO::Socket;
use Socket;

use constant USE_POSIX => defined eval 'use POSIX ":sys_wait_h"; 1';	## no critic

use constant DEFAULT_ADDR => qw(::1 127.0.0.1);
use constant DEFAULT_PORT => 5353;

use constant PACKETSZ => 512;

use constant SOCKOPT => eval {
	my @sockopt;
	push @sockopt, eval '[SOL_SOCKET, SO_REUSEADDR]';	## no critic
	push @sockopt, eval '[SOL_SOCKET, SO_REUSEPORT]';	## no critic

	my $pretest = sub {					# check that options safe to use
		return eval { IO::Socket::IP->new( Proto => "udp", Sockopts => [shift], Type => SOCK_DGRAM ) }
	};
	return grep { &$pretest($_) } @sockopt;			# without any guarantee that they work!
};



sub new {
	my ( $class, %self ) = @_;
	my $self = bless \%self, $class;

	$self->ReadZoneFile( $self{ZoneFile} ) if exists $self{ZoneFile};

	croak 'No reply handler!' unless ref( $self{ReplyHandler} ) eq "CODE";

	# local server addresses need to be accepted by a resolver
	my $LocalAddr = $self{LocalAddr} || [DEFAULT_ADDR];
	my $resolver  = Net::DNS::Resolver->new( nameservers => $LocalAddr );
	$resolver->force_v4(1) if $self{Force_IPv4};
	$resolver->force_v6(1) if $self{Force_IPv6};
	$self{LocalAddr} = [$resolver->nameservers];
	$self{LocalPort} ||= DEFAULT_PORT;

	$self{Truncate}	   = 1	 unless defined( $self{Truncate} );
	$self{IdleTimeout} = 120 unless defined( $self{IdleTimeout} );

	return $self;
}



sub ReadZoneFile {
	my ( $self, $file ) = @_;
	my $zonefile = Net::DNS::ZoneFile->new($file);

	my $RRhash = $self->{index} = {};
	my $RRlist = [];
	my @zonelist;
	while ( my $rr = $zonefile->read ) {
		push @{$RRhash->{lc $rr->owner}}, $rr;

		# Warning: Nasty trick abusing SOA to reference zone RR list
		if ( $rr->type eq 'SOA' ) {
			$RRlist = $rr->{RRlist} = [];
			push @zonelist, lc $rr->owner;
		} else {
			push @$RRlist, $rr;
		}
	}

	$self->{namelist}     = [sort { length($b) <=> length($a) } keys %$RRhash];
	$self->{zonelist}     = [sort { length($b) <=> length($a) } @zonelist];
	$self->{ReplyHandler} = sub { $self->ReplyHandler(@_) };
	return;
}



sub ReplyHandler {
	my ( $self, $qname, $qclass, $qtype, $peerhost, $query, $conn ) = @_;
	my $RRhash = $self->{index};
	my $rcode;
	my %headermask;
	my @ans;
	my @auth;

	if ( $qtype eq 'AXFR' ) {
		my $RRlist = $RRhash->{lc $qname} || [];
		my ($soa) = grep { $_->type eq 'SOA' } @$RRlist;
		if ($soa) {
			$rcode = 'NOERROR';
			push @ans, $soa, @{$soa->{RRlist}}, $soa;
		} else {
			$rcode = 'NOTAUTH';
		}
		return ( $rcode, \@ans, [], [], {}, {} );
	}

	my @RRname = @{$self->{namelist}};			# pre-sorted, longest first
	{
		my $RRlist = $RRhash->{lc $qname} || [];	# hash, then linear search
		my @match  = @$RRlist;				# assume $qclass always 'IN'
		if ( scalar(@match) ) {				# exact match
			$rcode = 'NOERROR';
		} elsif ( grep {/\.$qname$/i} @RRname ) {	# empty non-terminal
			$rcode = 'NOERROR';			# [NODATA]
		} else {
			$rcode = 'NXDOMAIN';
			foreach ( grep {/^[*][.]/} @RRname ) {
				my $wildcard = $_;		# match wildcard per RFC4592
				s/^\*//;			# delete leading asterisk
				s/([.?*+])/\\$1/g;		# escape dots and regex quantifiers
				next unless $qname =~ /[.]?([^.]+$_)$/i;
				my $encloser = $1;		# check no ENT encloses qname
				$rcode = 'NOERROR';
				last if grep {/(^|\.)$encloser$/i} @RRname;    # [NODATA]

				my ($q) = $query->question;	# synthesise RR at qname
				foreach my $rr ( @{$RRhash->{$wildcard}} ) {
					my $clone = bless( {%$rr}, ref($rr) );
					$clone->{owner} = $q->{qname};
					push @match, $clone;
				}
				last;
			}
		}
		push @ans, my @cname = grep { $_->type eq 'CNAME' } @match;
		$qname = $_->cname for @cname;
		redo if @cname;
		push @ans, @match if $qtype eq 'ANY';		# traditional, now out of favour
		push @ans, grep { $_->type eq $qtype } @match;
		unless (@ans) {
			foreach ( @{$self->{zonelist}} ) {
				s/([.?*+])/\\$1/g;		# escape dots and regex quantifiers
				next unless $qname =~ /[^.]+[.]$_[.]?$/i;
				push @auth, grep { $_->type eq 'SOA' } @{$RRhash->{$_}};
				last;
			}
		}
		$headermask{aa} = 1;
	}
	return ( $rcode, \@ans, \@auth, [], \%headermask, {} );
}



sub make_reply {
	my ( $self, $query, $sock ) = @_;
	my $verbose = $self->{Verbose};

	unless ($query) {
		my $empty = Net::DNS::Packet->new();		# create empty reply packet
		my $reply = $empty->reply();
		$reply->header->rcode("FORMERR");
		return $reply;
	}

	if ( $query->header->qr() ) {
		print "ERROR: invalid packet (qr set), dropping\n" if $verbose;
		return;
	}

	my $reply  = $query->reply();
	my $header = $reply->header;
	my $headermask;
	my $optionmask;

	my $opcode  = $query->header->opcode;
	my $qdcount = $query->header->qdcount;

	unless ($qdcount) {
		$header->rcode("NOERROR");

	} elsif ( $qdcount > 1 ) {
		$header->rcode("FORMERR");

	} else {
		my ($qr)   = $query->question;
		my $qname  = $qr->qname;
		my $qtype  = $qr->qtype;
		my $qclass = $qr->qclass;

		print $qr->string, "\n" if $verbose;

		my $conn = {
			peerhost => my $peer = $sock->peerhost,
			peerport => $sock->peerport,
			protocol => $sock->protocol,
			sockhost => $sock->sockhost,
			sockport => $sock->sockport
			};

		my ( $rcode, $ans, $auth, $add );
		my @arglist = ( $qname, $qclass, $qtype, $peer, $query, $conn );

		if ( $opcode eq "QUERY" ) {
			( $rcode, $ans, $auth, $add, $headermask, $optionmask ) =
					&{$self->{ReplyHandler}}(@arglist);

		} elsif ( $opcode eq "NOTIFY" ) {		#RFC1996
			if ( ref $self->{NotifyHandler} eq "CODE" ) {
				( $rcode, $ans, $auth, $add, $headermask, $optionmask ) =
						&{$self->{NotifyHandler}}(@arglist);
			} else {
				$rcode = "NOTIMP";
			}

		} elsif ( $opcode eq "UPDATE" ) {		#RFC2136
			if ( ref $self->{UpdateHandler} eq "CODE" ) {
				( $rcode, $ans, $auth, $add, $headermask, $optionmask ) =
						&{$self->{UpdateHandler}}(@arglist);
			} else {
				$rcode = "NOTIMP";
			}

		} else {
			print "ERROR: opcode $opcode unsupported\n" if $verbose;
			$rcode = "FORMERR";
		}

		if ( !defined($rcode) ) {
			print "remaining silent\n" if $verbose;
			return;
		}

		$header->rcode($rcode);

		push @{$reply->{answer}},     @$ans  if $ans;
		push @{$reply->{authority}},  @$auth if $auth;
		push @{$reply->{additional}}, @$add  if $add;
	}

	while ( my ( $key, $value ) = each %{$headermask || {}} ) {
		$header->$key($value);
	}

	while ( my ( $option, $value ) = each %{$optionmask || {}} ) {
		$reply->edns->option( $option, $value );
	}

	$header->print if $verbose && ( $headermask || $optionmask );

	return $reply;
}



sub TCP_connection {
	my ( $self, $socket ) = @_;
	my $timeout = $self->{IdleTimeout};
	my $verbose = $self->{Verbose};

	while (1) {
		alarm $timeout;
		my $l = '';
		my $n = sysread( $socket, $l, 2 );
		unless ( defined $n ) {
			redo if $!{EINTR};	## retry if aborted by signal
			die "sysread: $!";
		}
		last if $n == 0;
		my $msglen = unpack 'n', $l;

		my $buffer = '';
		while ( $msglen > ( my $l = length $buffer ) ) {
			my $n = sysread( $socket, $buffer, ( $msglen - $l ), $l );
			unless ( defined $n ) {
				redo if $!{EINTR};	## retry if aborted by signal
				die "sysread: $!";
			}
		}

		if ($verbose) {
			my $peer = $socket->peerhost;
			my $port = $socket->peerport;
			my $size = length $buffer;
			print "Received $size octets from [$peer] port $port\n";
		}

		my $query = Net::DNS::Packet->new( \$buffer );
		if ($@) {
			print "Error decoding query packet: $@\n" if $verbose;
			undef $query;	## force FORMERR reply
		}

		my $reply = $self->make_reply( $query, $socket );
		die 'Failed to create reply' unless defined $reply;

		my $segment = $reply->data(65500);		# limit to one TCP envelope
		my $length  = length $segment;
		warn "Multi-packet TCP response not implemented" if $reply->header->tc;
		if ($verbose) {
			print "TCP response (2 + $length octets) - ";
			print $socket->send( pack 'na*', $length, $segment ) ? "sent" : "failed: $!", "\n";
		} else {
			$socket->send( pack 'na*', $length, $segment );
		}
	}
	alarm 0;
	close $socket;
	return;
}



sub UDP_connection {
	my ( $self, $socket ) = @_;
	my $verbose = $self->{Verbose};
	alarm 3;

	my $buffer = "";
	return unless defined $socket->recv( $buffer, PACKETSZ );

	if ($verbose) {
		my $peer = $socket->peerhost;
		my $port = $socket->peerport;
		my $size = length $buffer;
		print "Received $size octets from [$peer] port $port\n";
	}

	my $query = Net::DNS::Packet->new( \$buffer );
	if ($@) {
		print "Error decoding query packet: $@\n" if $verbose;
		undef $query;		## force FORMERR reply
	}

	my $reply = $self->make_reply( $query, $socket );
	die 'Failed to create reply' unless defined $reply;

	my @UDPsize = ( $query && $self->{Truncate} ) ? $query->edns->UDPsize : ();
	if ($verbose) {
		my $response = $reply->data(@UDPsize);
		print 'UDP response (', length($response), ' octets) - ';
		print $socket->send($response) ? "sent" : "failed: $!", "\n";
	} else {
		$socket->send( $reply->data(@UDPsize) );
	}
	alarm 0;
	close $socket;
	return;
}



use constant DEBUG => 0;

sub logmsg { return print "$0 $$: @_ at ", scalar localtime(), "\n" }

sub spawn {
	my $coderef = shift;
	confess "usage: spawn CODEREF" unless ref($coderef) eq 'CODE';

	unless ( defined( my $pid = fork() ) ) {
		die "cannot fork: $!";
	} elsif ($pid) {
		logmsg "begat $pid" if DEBUG;
		return $pid;		## parent
	}

	# else ...
	$coderef->();			## child
	exit;
}

sub reaper {
	local $!;			## protect current error
	$SIG{CHLD} = \&reaper;		## no critic	sysV semantics
	while ( ( my $pid = waitpid( -1, USE_POSIX ? WNOHANG : 0 ) ) > 0 ) {
		logmsg "reaped $pid" if DEBUG;
	}
	return;
}


sub start_server {
	my $self = shift;
	my $list = $self->{LocalAddr};
	my $port = $self->{LocalPort};
	my @pid;
	foreach my $ip (@$list) {
		push @pid, spawn( sub { $self->TCP_server( $ip, $port ) } );
		push @pid, spawn( sub { $self->UDP_server( $ip, $port ) } );
	}
	return @pid;
}

sub start_noloop {
	my ( $self, $timeout ) = ( @_, 600 );
	my $list = $self->{LocalAddr};
	my $port = $self->{LocalPort};
	foreach my $ip (@$list) {
		spawn(	sub {
				alarm $timeout;
				$self->TCP_initialise( $ip, $port );
			} );
		spawn(	sub {
				alarm $timeout;
				$self->UDP_initialise( $ip, $port );
			} );
	}
	return;
}


sub TCP_initialise {
	my ( $self, $ip, $port ) = @_;
	my $socket = IO::Socket::IP->new(
		LocalAddr => $ip,
		LocalPort => $port,
		Sockopt	  => [SOCKOPT],
		Proto	  => "tcp",
		Listen	  => SOMAXCONN,
		Type	  => SOCK_STREAM
		)
			|| die "can't setup TCP socket";

	logmsg "TCP server [$ip] port $port started";

	{
		my $client = $socket->accept() || do {
			redo if $!{EINTR};	## retry if aborted by signal
			die "accept: $!";
		};

		spawn( sub { $self->TCP_connection($client) } );
	}
	return $socket;
}

sub TCP_server {
	my ( $self, $ip, $port ) = @_;
	my $socket = $self->TCP_initialise( $ip, $port );
	while (1) {
		my $client = $socket->accept() || do {
			redo if $!{EINTR};	## retry if aborted by signal
			die "accept: $!";
		};

		spawn( sub { $self->TCP_connection($client) } );
	}
	exit;
}


sub UDP_initialise {
	my ( $self, $ip, $port ) = @_;
	my $socket = IO::Socket::IP->new(
		LocalAddr => $ip,
		LocalPort => $port,
		Sockopt	  => [SOCKOPT],
		Proto	  => "udp",
		Type	  => SOCK_DGRAM
		)
			|| die "can't setup UDP socket";

	logmsg "UDP server [$ip] port $port started";

	my $select = IO::Select->new($socket);
	{
		local $! = 0;
		scalar( my @ready = $select->can_read() ) || do {
			redo if $!{EINTR};	## retry if aborted by signal
			die "select->can_read(): $!";
		};

		foreach my $client (@ready) {
			spawn( sub { $self->UDP_connection($client) } );
		}
	}
	return $socket;
}

sub UDP_server {
	my ( $self, $ip, $port ) = @_;
	my $socket = $self->UDP_initialise( $ip, $port );
	my $select = IO::Select->new($socket);
	while (1) {
		local $! = 0;
		scalar( my @ready = $select->can_read() ) || do {
			redo if $!{EINTR};	## retry if aborted by signal
			die "select->can_read(): $!";
		};

		foreach my $client (@ready) {
			spawn( sub { $self->UDP_connection($client) } );
		}
	}
	exit;
}



sub main_loop {
	local $SIG{CHLD} = \&reaper;
	my @pid = shift->start_server;
	local $SIG{TERM} = sub { kill( 'TERM', @pid ) };
	1 while waitpid( -1, 0 ) > 0;
	logmsg "@pid terminated";
	return;
}



sub loop_once {
	my ( $self, @timeout ) = @_;
	local $SIG{CHLD} = \&reaper;
	$self->start_noloop(@timeout);
	1 while waitpid( -1, 0 ) > 0;	## park main process until timeout or
	return;				## user CTRL_C kills remaining children
}


1;
__END__



__END__

